/*
Quadrature decoder using PIO1_SM0 to decode PHASE_A and PHASE_B signals at GPIO10 and GPIO11.
A preload value can be set (X-reg of PIO1_SM0) and is loaded at '0' state of INDEX_PIN GPIO12.

Each rising / falling edge is counted up to 10e7 counts/s. Variable 'q_cnt0' is updated via
DMA-channel CH8. After max. transfer-count of CH8, DMA-CH9 is used to retrigger CH8.

2024-03-17
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rp2040.h"

// inputs channel 0
#define PHASE_A0       10
#define PHASE_B0       11
#define INDEX_PIN0     12


// unconditional JMP = INSTR_MEM_adr
#define JMP_DEC       22
#define JMP_LOOP      15
#define JMP_LOOPx     18
#define JMP_TEST      17
#define JMP_INC       24
#define JMP_INC_CONT  26
#define JMP_OUT       27
#define JMP_PRELOAD   29
#define JMP_OUTx      30

#define DMA_COUNT     0xFFFFFFFF

int q_cnt0, q_cnt0_offset;     
static uint32_t dma_dummy0;


void init_PIO1(void)
{
  uint32_t *prg;
  static int pio1_init;
  if(pio1_init) return;                          // init already done
  pio1_init = 1;
// enable PIO1 
  RESETS_CLR->RESET = RESETS_RESET_pio1_Msk;    
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio1_Msk));
  
// enable GPIOs
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));

    prg = (uint32_t *)&PIO1->INSTR_MEM0;        // pgm_start

// 'xx' is new state
// 00xx state
    *prg++ = JMP_LOOP;
    *prg++ = JMP_DEC;
    *prg++ = JMP_INC;
    *prg++ = JMP_LOOP;

// 01xx state
    *prg++ = JMP_INC;
    *prg++ = JMP_LOOP;
    *prg++ = JMP_LOOP;
    *prg++ = JMP_DEC;

// 10xx state
    *prg++ = JMP_DEC;
    *prg++ = JMP_LOOP;
    *prg++ = JMP_LOOP;
    *prg++ = JMP_INC;

// 11xx state
    *prg++ = JMP_LOOP;
    *prg++ = JMP_INC;
    *prg++ = JMP_DEC;
    
// ADR15:      
// LOOP: test reset if X != Y
    *prg++ = JMP_TEST | 0xa0;     // JMP X != Y
    *prg++ = JMP_LOOPx;           // X == Y, don't test again
// ADR17:
// TEST: Reset, if INDEX_PIN1 == 1     
    *prg++ = JMP_PRELOAD | 0xc0;  // JMP PIN == 1, reset
    
// ADR18:      
// LOOPx: normal loop without reset
// JMP-adr calculation
    *prg++ = 0x60c2;              // OUT ISR, 2     get old state to 'xx'
    *prg++ = 0x4002;              // IN PINS, 2     new state = 'xx' << 2 + 'ba'
// OSR temp-reg for ISR
    *prg++ = 0xa0e6;              // MOV OSR, ISR   save act. state 'xxba' 
    *prg++ = 0xa0a6;              // MOV PC, ISR    jump to function
    
// ADR22:
// DEC: decrement Y and output
    *prg++ = JMP_DEC+1 + 0x80;    // JMP Y--        and continue
    *prg++ = JMP_OUT;             // JMP OUT

// ADR24:
// INC: increment Y 
    *prg++ = 0xa04a;              // MOV Y, ~Y      invert
    *prg++ = JMP_INC_CONT | 0x80; // JMP Y--,       Y-- and dummy JMP to INC_CONT
// INC_CONT:
    *prg++ = 0xa04a;              // MOV Y, ~Y      invert again
    
// ADR27:
// OUT: output Y via RXFIFO if no reset
    *prg++ = JMP_PRELOAD | 0xc0;  // JMP PIN == 1, reset
    *prg++ = JMP_OUTx;            // no reset
// ADR29:
// PRELOAD: output X if reset
    *prg++ = 0xa041;              // MOV Y, X     Y = X, preload value
// ADR30:
// OUTx: output Y via RXFIFO      
    *prg++ = 0x4040;              // MOV ISR, Y     copy to ISR
// ADR31: wrapping to LOOP
    *prg++ = 0x8000;              // PUSH noblock   send without blocking, recieve via DMA
}

void init_PIO1_SM0(void)
{
  init_PIO1();
// PhaseA: GPIO10, PhaseB: GPIO11, Index: GPIO12
  IO_BANK0->GPIO10_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO11_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  PADS_BANK0->GPIO10 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  PADS_BANK0->GPIO11 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  IO_BANK0_SET->GPIO12_CTRL = 1 << IO_BANK0_GPIO0_CTRL_INOVER_Pos; // '0' active
  PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
// set input group starting @GPIO10
  PIO1->SM0_PINCTRL =  PHASE_A0 << PIO0_SM0_PINCTRL_IN_BASE_Pos;       // 10 + 11
  PIO1->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                       (JMP_LOOP << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                       (INDEX_PIN0  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos); // index-pin0: 12
  
     
  PIO1->SM0_SHIFTCTRL = 0 << PIO0_SM0_SHIFTCTRL_IN_SHIFTDIR_Pos | // IN shift left
                        1 << PIO0_SM0_SHIFTCTRL_OUT_SHIFTDIR_Pos; // OUT shift right
  PIO1->SM0_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos;                // default, max. speed
  PIO1->SM0_ADDR = JMP_OUT;
  PIO1_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));              // enable PIO1-SM0

}


// if needed all interrupts should be disabled
void preload_qcnt0_val(int x)
{
  PIO1_CLR->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));      // stop PIO1-SM0
  PIO1->TXF0 = x;                         // write new value
  PIO1->SM0_INSTR = 0x8080;               // PULL OSR
  PIO1->SM0_INSTR = 0xa027;               // MOV X, OSR
  PIO1->SM0_INSTR = 0xa041;               // MOV Y, X     Y = X, preload value
  PIO1_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));      // start PIO1-SM0
}  
 

// no ISR needed: get 'q_cnt0' via DMA->CH8
// reload CH8_TRANS_COUNT via DMA->CH9
void init_qcnt0(int preload)
{
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// regular transfer  
  DMA->CH8_READ_ADDR = (uint32_t) &PIO1->RXF0;
  DMA->CH8_WRITE_ADDR = (uint32_t) &q_cnt0;
  DMA->CH8_TRANS_COUNT = DMA_COUNT;
  DMA->CH8_CTRL_TRIG =  DREQ_PIO1_RX0 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // PIO1-RXF0
                        9 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // trigger CH9
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |       // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                    // enable
      
// retrigger CH8_TRANS_COUNT
  DMA->CH9_READ_ADDR =  (uint32_t) &dma_dummy0;                      // dummy location
  DMA->CH9_WRITE_ADDR = (uint32_t) &dma_dummy0;
  DMA->CH9_TRANS_COUNT = 1;                                          // only 1 reload-transfer
  DMA->CH9_CTRL_TRIG = DREQ_AUTO << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // free running
                        8 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // retrigger CH8  
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |       // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                    // enable
  init_PIO1_SM0();
  preload_qcnt0_val(preload);
}

// get raw data
int get_qcnt0(void)
{
  return(q_cnt0);
}

// read rel. data
int read_qcnt0(void)
{
  return(q_cnt0 - q_cnt0_offset);
}

// preset act. position to x
void preset_qcnt0(int x)
{
  q_cnt0_offset = -q_cnt0 + x;
}

// set active state of INDEX_PIN0
void set_index0_pol(int i)
{
  if(i) {                                                       
    PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pulldown + schmitttrigger
                      PADS_BANK0_GPIO0_PDE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
// normal INDEX_PIN0 polarity: '1' active
    IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0; 

  } else {                                                      
    PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
// invert INDEX_PIN0 polarity: '0' active
    IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_INOVER_INVERT << IO_BANK0_GPIO0_CTRL_INOVER_Pos |
                          IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;
  }                      
}

