/*
Quadrature decoder using PIO1_SM1 to decode PHASE_A and PHASE_B signals at GPIO14 and GPIO15.
A preload value can be set (X-reg of PIO1_SM1) and is loaded at '0' state of INDEX_PIN GPIO13.

Each rising / falling edge is counted up to 10e7 counts/s. Variable 'q_cnt1' is updated via
DMA-channel CH10. After max. transfer-count of CH10, DMA-CH11 is used to retrigger CH10.

2024-03-17
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rp2040.h"
#define DMA_COUNT     0xFFFFFFFF

// inputs channel 1
#define PHASE_A1        14
#define PHASE_B1        15
#define INDEX_PIN1      13
#define JMP_LOOP        15

int q_cnt1, q_cnt1_offset;     
static uint32_t dma_dummy1;

extern void init_PIO1(void);

// if needed all interrupts should be disabled
void preload_qcnt1_val(int x)
{
  PIO1_CLR->CTRL = (2 << (PIO0_CTRL_SM_ENABLE_Pos));      // stop PIO1-SM1
  PIO1->TXF1 = x;                         // write new value
  PIO1->SM1_INSTR = 0x8080;               // PULL OSR
  PIO1->SM1_INSTR = 0xa027;               // MOV X, OSR
  PIO1->SM1_INSTR = 0xa041;               // MOV Y, X     Y = X = preload value
  PIO1_SET->CTRL = (2 << (PIO0_CTRL_SM_ENABLE_Pos));      // start PIO1-SM1
}  


void init_PIO1_SM1(void)
{
  init_PIO1();
// PhaseA: GPIO14, PhaseB: GPIO15, Index: GPIO13
  IO_BANK0->GPIO14_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO15_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO13_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  PADS_BANK0->GPIO14 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  PADS_BANK0->GPIO15 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  IO_BANK0_SET->GPIO13_CTRL = 1 << IO_BANK0_GPIO0_CTRL_INOVER_Pos; // '0' active
  PADS_BANK0->GPIO13 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
// set input group starting @GPIO14
  PIO1->SM1_PINCTRL =  PHASE_A1 << PIO0_SM0_PINCTRL_IN_BASE_Pos;       // 14 + 15
  PIO1->SM1_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                       (JMP_LOOP << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                       (INDEX_PIN1  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos); // index-pin1: 13
  
     
  PIO1->SM1_SHIFTCTRL = 0 << PIO0_SM0_SHIFTCTRL_IN_SHIFTDIR_Pos | // IN shift left
                        1 << PIO0_SM0_SHIFTCTRL_OUT_SHIFTDIR_Pos; // OUT shift right
  PIO1->SM1_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos;                // default, max. speed
  PIO1_SET->CTRL = (2 << (PIO0_CTRL_SM_ENABLE_Pos));              // enable PIO1-SM1
}
 

// no ISR needed: get 'q_cnt1' via DMA->CH10
// reload CH10_TRANS_COUNT via DMA->CH11
void init_qcnt1(int preload)
{
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// regular transfer  
  DMA->CH10_READ_ADDR = (uint32_t) &PIO1->RXF1;
  DMA->CH10_WRITE_ADDR = (uint32_t) &q_cnt1;
  DMA->CH10_TRANS_COUNT = DMA_COUNT;
  DMA->CH10_CTRL_TRIG =  DREQ_PIO1_RX1 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // PIO1-RXF1
                        11 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // trigger CH11
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |        // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                     // enable
      
// retrigger CH10_TRANS_COUNT
  DMA->CH11_READ_ADDR =  (uint32_t) &dma_dummy1;                      // dummy location
  DMA->CH11_WRITE_ADDR = (uint32_t) &dma_dummy1;
  DMA->CH11_TRANS_COUNT = 1;                                          // only 1 reload-transfer
  DMA->CH11_CTRL_TRIG = DREQ_AUTO << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // free running
                        10 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // retrigger CH10  
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |        // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                     // enable
  init_PIO1_SM1();
  preload_qcnt1_val(preload);
}

// get raw data
int get_qcnt1(void)
{
  return(q_cnt1);
}

// read rel. data
int read_qcnt1(void)
{
  return(q_cnt1 - q_cnt1_offset);
}

// preset act. position to x
void preset_qcnt1(int x)
{
  q_cnt1_offset = -q_cnt1 + x;
}

// set active state of INDEX_PIN1
void set_index1_pol(int i)
{
  if(i) {                                                       
    PADS_BANK0->GPIO13 = PADS_BANK0_GPIO0_IE_Msk |              // pulldown + schmitttrigger
                      PADS_BANK0_GPIO0_PDE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
// normal INDEX_PIN1 polarity: '1' active
    IO_BANK0->GPIO13_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0; 

  } else {                                                      
    PADS_BANK0->GPIO13 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
// invert INDEX_PIN1 polarity: '0' active
    IO_BANK0->GPIO13_CTRL = IO_BANK0_GPIO0_CTRL_INOVER_INVERT << IO_BANK0_GPIO0_CTRL_INOVER_Pos |
                          IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;
  }                      
}
