/*
Quadrature decoder using PIO1_SM2 to decode PHASE_A and PHASE_B signals at GPIO16 and GPIO17.

Each rising / falling edge is counted up to 10e6 counts/s. Variable 'q_cnt2' is updated via
DMA-channel CH7. TRANS_COUNT is retriggerd each time q_cnt2 is read via get_qcnt2(). 

2024-03-17
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rp2040.h"

#define BIT(x) (1<<x)
#define DMA_COUNT       0xFFFFFFFF  // max. value
#define DIV_10          10          // 1/10 of max. PIO1_SM2 speed

// inputs channel 2
#define PHASE_A2        16
#define PHASE_B2        17
#define NO_INDEX_PIN    31          // dummy, will always be read as '0'
#define JMP_LOOP        15          // for wrapping

int q_cnt2, q_cnt2_offset;     

extern void init_PIO1(void);

void init_PIO1_SM2(void)
{
  init_PIO1();
// PhaseA: GPIO16, PhaseB: GPIO17
  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  PADS_BANK0->GPIO16 = PADS_BANK0_GPIO0_IE_Msk |                  // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  PADS_BANK0->GPIO17 = PADS_BANK0_GPIO0_IE_Msk |                  // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
// set input group starting @GPIO16
  PIO1->SM2_PINCTRL =  PHASE_A2 << PIO0_SM0_PINCTRL_IN_BASE_Pos;  // 16 + 17
  PIO1->SM2_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                       (JMP_LOOP << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                       (NO_INDEX_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos); // dummy
  
     
  PIO1->SM2_SHIFTCTRL = 0 << PIO0_SM0_SHIFTCTRL_IN_SHIFTDIR_Pos | // IN shift left
                        1 << PIO0_SM0_SHIFTCTRL_OUT_SHIFTDIR_Pos; // OUT shift right
  PIO1->SM2_CLKDIV = DIV_10 << PIO0_SM0_CLKDIV_INT_Pos;           // 1/10 of max. speed
  PIO1_SET->CTRL = (4 << (PIO0_CTRL_SM_ENABLE_Pos));              // enable PIO1-SM2
}
 

// no ISR needed: get 'q_cnt2' via DMA->CH7
// reload CH7_TRANS_COUNT by function get_qcnt2()
void init_qcnt2(int preload)                                      // preload handled as offset
{
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// regular transfer  
  DMA->CH7_READ_ADDR = (uint32_t) &PIO1->RXF2;
  DMA->CH7_WRITE_ADDR = (uint32_t) &q_cnt2;
  DMA->CH7_TRANS_COUNT = DMA_COUNT;
  DMA->CH7_CTRL_TRIG =  DREQ_PIO1_RX2 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // PIO1-RXF2
                        7 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |     // no trigger channel
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |    // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                 // enable
  q_cnt2_offset = -preload;  
  init_PIO1_SM2();
}

// get raw data
int get_qcnt2(void)
{
  DMA->CHAN_ABORT = BIT(7);                                       // abort CH7
  while(DMA->CHAN_ABORT);                                         // wait until done
  DMA_SET->CH7_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;              // start again
  return(q_cnt2);                                                 // current value
}

// read rel. data
int read_qcnt2(void)
{
  return(get_qcnt2() - q_cnt2_offset);
}

// preset act. position to x
void preset_qcnt2(int x)
{
  q_cnt2_offset = -get_qcnt2() + x;
}
